# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from openai.types.chat import ChatCompletionMessageParam
from mito_ai.completions.models import AgentResponse, AgentSmartDebugMetadata, MessageType, ResponseFormatInfo
from mito_ai.completions.prompt_builders.agent_smart_debug_prompt import create_agent_smart_debug_prompt
from mito_ai.completions.providers import OpenAIProvider
from mito_ai.completions.message_history import GlobalMessageHistory
from mito_ai.completions.completion_handlers.completion_handler import CompletionHandler
from mito_ai.completions.completion_handlers.open_ai_models import MESSAGE_TYPE_TO_MODEL
from mito_ai.completions.completion_handlers.utils import append_agent_system_message


__all__ = ["get_agent_auto_error_fixup_completion"]

class AgentAutoErrorFixupHandler(CompletionHandler[AgentSmartDebugMetadata]):
    """Handler for smart debug completions."""
    
    @staticmethod
    async def get_completion(
        metadata: AgentSmartDebugMetadata,
        provider: OpenAIProvider,
        message_history: GlobalMessageHistory
    ) -> str:
        """Get a smart debug completion from the AI provider."""

        # Add the system message if it doesn't already exist
        await append_agent_system_message(message_history, provider, metadata.threadId, metadata.isChromeBrowser)
        
        # Create the prompt
        prompt = create_agent_smart_debug_prompt(metadata)
        display_prompt = metadata.errorMessage
        
        # Add the prompt to the message history
        new_ai_optimized_message: ChatCompletionMessageParam = {"role": "user", "content": prompt}
        new_display_optimized_message: ChatCompletionMessageParam = {"role": "user", "content": display_prompt}
        await message_history.append_message(new_ai_optimized_message, new_display_optimized_message, provider, metadata.threadId)
        
        # Get the completion
        completion = await provider.request_completions(
            messages=message_history.get_ai_optimized_history(metadata.threadId), 
            model=MESSAGE_TYPE_TO_MODEL[MessageType.AGENT_AUTO_ERROR_FIXUP],
            response_format_info=ResponseFormatInfo(
                name='agent_response',
                format=AgentResponse
            ),
            message_type=MessageType.AGENT_AUTO_ERROR_FIXUP,
            user_input=metadata.errorMessage,
            thread_id=metadata.threadId
        )
        
        # Add the response to message history
        ai_response_message: ChatCompletionMessageParam = {"role": "assistant", "content": completion}
        display_response_message: ChatCompletionMessageParam = {"role": "assistant", "content": completion}
        await message_history.append_message(ai_response_message, display_response_message, provider, metadata.threadId)

        return completion

# Use the static method directly
get_agent_auto_error_fixup_completion = AgentAutoErrorFixupHandler.get_completion



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from mito_ai.completions.models import MessageType

# Model name constants
GPT4O_MINI = "gpt-4o-mini"
GPT4O = "gpt-4o"
O3_MINI = "o3-mini"
GPT41 = "gpt-4.1"
GPT41_NANO = "gpt-4.1-nano"

# Mapping from message type to model name
MESSAGE_TYPE_TO_MODEL = {
    MessageType.SMART_DEBUG: GPT41,
    MessageType.CHAT: GPT41,
    MessageType.CODE_EXPLAIN: GPT41,
    MessageType.INLINE_COMPLETION: GPT41_NANO,
    MessageType.AGENT_AUTO_ERROR_FIXUP: GPT41,
    MessageType.AGENT_EXECUTION: GPT41,
    MessageType.CHAT_NAME_GENERATION: GPT41_NANO
}



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from dataclasses import dataclass
import json
from typing import Any, Final
import tornado
import os
from jupyter_server.base.handlers import APIHandler
from mito_ai.rules.utils import RULES_DIR_PATH, get_all_rules, get_rule, set_rules_file


class RulesHandler(APIHandler):
    """Handler for operations on a specific setting"""
    
    @tornado.web.authenticated
    def get(self, key=None):
        """Get a specific rule by key or all rules if no key provided"""
        if key is None or key == '':
            # No key provided, return all rules
            rules = get_all_rules()
            self.finish(json.dumps(rules))
        else:
            # Key provided, return specific rule
            rule_content = get_rule(key)
            if rule_content is None:
                self.set_status(404)
                self.finish(json.dumps({"error": f"Rule with key '{key}' not found"}))
            else:
                self.finish(json.dumps({"key": key, "content": rule_content}))
    
    @tornado.web.authenticated
    def put(self, key):
        """Update or create a specific setting"""
        data = json.loads(self.request.body)
        if 'content' not in data:
            self.set_status(400)
            self.finish(json.dumps({"error": "Content is required"}))
            return
            
        set_rules_file(key, data['content'])
        self.finish(json.dumps({"status": "updated", "rules file ": key}))





# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import json
import anthropic
from typing import Dict, Any, Optional, Tuple, Union, Callable, List

from anthropic.types import Message, MessageParam, ToolUnionParam
from mito_ai.completions.models import ResponseFormatInfo, CompletionReply, CompletionStreamChunk, CompletionItem, AgentResponse, MessageType
from openai.types.chat import ChatCompletionMessageParam
from mito_ai.utils.anthropic_utils import get_anthropic_completion_from_mito_server, stream_anthropic_completion_from_mito_server, get_anthropic_completion_function_params

MAX_TOKENS = 2_000

def _extract_and_parse_json_response(response: Message) -> Union[object, Any]:
    """
    Extracts and parses the JSON response from the Claude API.
    """
    try:
        # Check for tool use in the response
        for content_block in response.content:
            if content_block.type == "tool_use" and content_block.name == "agent_response":
                result = content_block.input
                return result

        # If no tool use was found, try to parse the text response
        text_response = None
        for content_block in response.content:
            if content_block.type == "text":
                text_response = content_block.text
                break

        if text_response:
            # Try to extract JSON from the text response
            import re
            json_pattern = r'(\{.*\})'
            match = re.search(json_pattern, text_response, re.DOTALL)
            if match:
                try:
                    json_response = json.loads(match.group(0))
                    return json_response
                except json.JSONDecodeError:
                    pass

        raise Exception("No valid AgentResponse format found in the response")
    except Exception as e:
        raise Exception(f"Failed to parse response: {e}")


def _get_system_prompt_and_messages(messages: List[ChatCompletionMessageParam]) -> Tuple[Union[str, anthropic.NotGiven], List[MessageParam]]:
    """
    Convert a list of OpenAI messages to a list of Anthropic messages.
    """
    anthropic_messages: List[MessageParam] = []
    system_prompt: Union[str, anthropic.NotGiven] = anthropic.NotGiven()
    
    for message in messages:
        if 'content' not in message:
            continue
        
        # We assume that the converastion only has one system message.
        # Or if there are multiple, we take the last one.
        if message['role'] == 'system':
            system_prompt = str(message['content'])

        # Construct the messages for the user and assistant in Anthropic format.
        if message['role'] == 'user':
            anthropic_messages.append(MessageParam(role='user', content=str(message['content'])))
        elif message['role'] == 'assistant':
            anthropic_messages.append(MessageParam(role='assistant', content=str(message['content'])))

    return system_prompt, anthropic_messages


class AnthropicClient:
    """
    A client for interacting with the Anthropic API or the Mito server fallback.
    """

    def __init__(self, api_key: Optional[str], model: str, timeout: int = 30, max_retries: int = 1):
        self.api_key = api_key
        self.model = model
        self.timeout = timeout
        self.max_retries = max_retries
        if api_key:
            self.client = anthropic.Anthropic(api_key=api_key)
        else:
            self.client = None  # type: ignore[assignment]

    async def request_completions(self, messages: List[ChatCompletionMessageParam],
                                  response_format_info: Optional[ResponseFormatInfo] = None,
                                  message_type: MessageType = MessageType.CHAT) -> Any:
        """
        Get a response from Claude or the Mito server that adheres to the AgentResponse format.
        """
        try:
            anthropic_system_prompt, anthropic_messages = _get_system_prompt_and_messages(messages)
            # Build provider_data once
            tools: Optional[List[ToolUnionParam]] = None
            tool_choice = None
            if response_format_info and response_format_info.name == "agent_response":
                tools = [{
                    "name": "agent_response",
                    "description": "Output a structured response following the AgentResponse format",
                    "input_schema": AgentResponse.model_json_schema()
                }]
                tool_choice = {"type": "tool", "name": "agent_response"}
            provider_data = get_anthropic_completion_function_params(
                model=self.model,
                messages=anthropic_messages,
                max_tokens=MAX_TOKENS,
                temperature=0,
                system=anthropic_system_prompt,
                tools=tools,
                tool_choice=tool_choice,
                stream=None,
                response_format_info=response_format_info
            )
            if self.api_key:
                # Unpack provider_data for direct API call
                assert self.client is not None
                response = self.client.messages.create(**provider_data)
                if tools:
                    result = _extract_and_parse_json_response(response)
                    return json.dumps(result) if not isinstance(result, str) else result
                else:
                    content = response.content
                    if content[0].type == "text":
                        return content[0].text
                    else:
                        return ""
            else:
                # Only pass provider_data to the server
                
                system_val = provider_data.get("system", None)
                if system_val is not None and system_val is not anthropic.NotGiven:
                    system_tuple = (system_val,)
                else:
                    system_tuple = tuple()  # type: ignore[assignment]

                response = await get_anthropic_completion_from_mito_server(
                    model=provider_data["model"],
                    max_tokens=provider_data["max_tokens"],
                    temperature=provider_data["temperature"],
                    system=system_tuple,
                    messages=provider_data["messages"],
                    tools=provider_data.get("tools"),
                    tool_choice=provider_data.get("tool_choice"),
                    message_type=message_type
                )
                return response
        except anthropic.RateLimitError:
            raise Exception("Rate limit exceeded. Please try again later or reduce your request frequency.")
        except Exception as e:
            return f"Error streaming content: {str(e)}"

    async def stream_response(self, messages: List[ChatCompletionMessageParam], message_id: str, message_type: MessageType,
                              reply_fn: Callable[[Union[CompletionReply, CompletionStreamChunk]], None]) -> str:
        try:
            anthropic_system_prompt, anthropic_messages = _get_system_prompt_and_messages(messages)
            accumulated_response = ""

            if self.api_key:
                assert self.client is not None
                stream = self.client.messages.create(
                    model=self.model,
                    max_tokens=MAX_TOKENS,
                    temperature=0,
                    system=anthropic_system_prompt,
                    messages=anthropic_messages,
                    stream=True
                )


                for chunk in stream:
                    if chunk.type == "content_block_delta" and chunk.delta.type == "text_delta":
                        content = chunk.delta.text
                        accumulated_response += content

                        is_finished = chunk.type == "message_stop"

                        reply_fn(CompletionStreamChunk(
                            parent_id=message_id,
                            chunk=CompletionItem(
                                content=content,
                                isIncomplete=not is_finished,
                                token=message_id,
                            ),
                            done=is_finished,
                        ))

            else:
                async for stram_chunk in stream_anthropic_completion_from_mito_server(
                    model=self.model,
                    max_tokens=MAX_TOKENS,
                    temperature=0,
                    system=(anthropic_system_prompt,),
                    messages=anthropic_messages,
                    stream=True,
                    message_type=message_type
                ):
                    accumulated_response += stram_chunk
                    reply_fn(CompletionStreamChunk(
                        parent_id=message_id,
                        chunk=CompletionItem(
                            content=stram_chunk,
                            isIncomplete=True,
                            token=message_id,
                        ),
                        done=False,
                    ))

            return accumulated_response

        except anthropic.RateLimitError:
            raise Exception("Rate limit exceeded. Please try again later or reduce your request frequency.")

        except Exception as e:
            return f"Error streaming content: {str(e)}"





# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from __future__ import annotations
from typing import Any, AsyncGenerator, Callable, Dict, List, Optional, Union, Type

from openai import api_key

from mito_ai import constants
from openai.types.chat import ChatCompletionMessageParam
from traitlets import Instance, Unicode, default, validate
from traitlets.config import LoggingConfigurable
from openai.types.chat import ChatCompletionMessageParam

from mito_ai import constants
from mito_ai.enterprise.utils import is_azure_openai_configured
from mito_ai.gemini_client import GeminiClient
from mito_ai.openai_client import OpenAIClient
from mito_ai.anthropic_client import AnthropicClient
from mito_ai.logger import get_logger
from mito_ai.completions.models import (
    AICapabilities,
    CompletionError,
    CompletionItem,
    CompletionItemError,
    CompletionReply,
    CompletionStreamChunk,
    MessageType,
    ResponseFormatInfo,
)
from mito_ai.utils.telemetry_utils import (
    KEY_TYPE_PARAM,
    MITO_AI_COMPLETION_ERROR,
    MITO_SERVER_KEY,
    USER_KEY,
    log,
    log_ai_completion_success,
)

__all__ = ["OpenAIProvider"]


class OpenAIProvider(LoggingConfigurable):
    """Provide AI feature through OpenAI services."""

    api_key = Unicode(
        config=True,
        allow_none=True,
        help="OpenAI API key. Default value is read from the OPENAI_API_KEY environment variable.",
    )

    last_error = Instance(
        CompletionError,
        allow_none=True,
        help="""Last error encountered when using the OpenAI provider.

This attribute is observed by the websocket provider to push the error to the client.""",
    )

    def __init__(self, **kwargs: Dict[str, Any]) -> None:
        # Create config for OpenAI client before parent initialization
        config = kwargs.get('config', {})
        if 'api_key' in kwargs:
            config['OpenAIClient'] = {'api_key': kwargs['api_key']}
        kwargs['config'] = config

        super().__init__(log=get_logger(), **kwargs)
        self.last_error = None
        self._openai_client: Optional[OpenAIClient] = None
        self._gemini_client: Optional[GeminiClient] = None
        self._anthropic_client: Optional[AnthropicClient] = None

        # Initialize OpenAI client with the configured api_key
        self._openai_client = OpenAIClient(parent=self)

        # Initialize Gemini client if configured
        if constants.GEMINI_MODEL and constants.GEMINI_API_KEY:
            self._gemini_client = GeminiClient(
                api_key=constants.GEMINI_API_KEY,
                model=constants.GEMINI_MODEL
            )

        # Initialize Anthropic client if configured
        if constants.CLAUDE_MODEL and constants.CLAUDE_API_KEY:
            self._anthropic_client = AnthropicClient(
                api_key=constants.CLAUDE_API_KEY,
                model=constants.CLAUDE_MODEL
            )

    @property
    def capabilities(self) -> AICapabilities:
        """Get the provider capabilities."""
        if constants.CLAUDE_MODEL and constants.CLAUDE_API_KEY and not self.api_key:
            return AICapabilities(
                configuration={
                    "model": constants.CLAUDE_MODEL
                },
                provider="Claude",
            )

        if constants.GEMINI_MODEL and constants.GEMINI_API_KEY and not self.api_key:
            return AICapabilities(
                configuration={
                    "model": constants.GEMINI_MODEL
                },
                provider="Gemini",
            )

        if self._openai_client:
            return self._openai_client.capabilities

        return AICapabilities(
            configuration={
                "model": "gpt-4.1",
            },
            provider="Mito server",
        )

    @property
    def key_type(self) -> str:
        """Returns the authentication key type being used."""
        if constants.CLAUDE_MODEL and constants.CLAUDE_API_KEY and not self.api_key:
            return "claude"

        if constants.GEMINI_MODEL and constants.GEMINI_API_KEY and not self.api_key:
            return "gemini"

        if self._openai_client:
            return self._openai_client.key_type

        return MITO_SERVER_KEY

    async def request_completions(
            self,
            message_type: MessageType,
            messages: List[ChatCompletionMessageParam],
            model: str,
            response_format_info: Optional[ResponseFormatInfo] = None,
            user_input: Optional[str] = None,
            thread_id: Optional[str] = None
    ) -> str:
        """
        Request completions from the AI provider.

        Args:
            message_type: The type of message to request completions for.
            messages: The messages to request completions for.
            model: The model to request completions for.
        Returns:
            The completion from the AI provider.
        """
        # Reset the last error
        self.last_error = None
        completion = None
        last_message_content = str(messages[-1].get('content', '')) if messages else ""

        try:
            # Handle Claude API calls
            if constants.CLAUDE_MODEL and not self.api_key:
                if self._anthropic_client is None:
                    if constants.CLAUDE_API_KEY:
                        self._anthropic_client = AnthropicClient(
                            api_key=constants.CLAUDE_API_KEY,
                            model=constants.CLAUDE_MODEL
                        )
                    else:
                        self._anthropic_client = AnthropicClient(
                            api_key=None,
                            model=constants.CLAUDE_MODEL
                        )
                completion = await self._anthropic_client.request_completions(messages, response_format_info, message_type)

            # Handle Gemini API calls
            elif constants.GEMINI_MODEL and not self.api_key:
                if self._gemini_client is None:
                    if constants.GEMINI_API_KEY:
                        self._gemini_client = GeminiClient(
                            api_key=constants.GEMINI_API_KEY,
                            model=constants.GEMINI_MODEL
                        )
                    else:
                        self._gemini_client = GeminiClient(
                            api_key=None,
                            model=constants.GEMINI_MODEL
                        )
                completion = await self._gemini_client.request_completions(messages, response_format_info, message_type)

            # Handle OpenAI and other providers
            elif self._openai_client:
                completion = await self._openai_client.request_completions(
                    message_type=message_type,
                    messages=messages,
                    model=model,
                    response_format_info=response_format_info
                )
            else:
                raise ValueError("No AI provider configured")

            # Log the successful completion
            log_ai_completion_success(
                key_type=USER_KEY if self.key_type == "user" else MITO_SERVER_KEY,
                message_type=message_type,
                last_message_content=last_message_content,
                response={"completion": completion},
                user_input=user_input or "",
                thread_id=thread_id or ""
            )

            return completion

        except BaseException as e:
            self.log.exception(f"Error during request_completions: {e}")
            self.last_error = CompletionError.from_exception(e)
            log(MITO_AI_COMPLETION_ERROR, params={KEY_TYPE_PARAM: self.key_type}, error=e)
            raise

    async def stream_completions(
            self,
            message_type: MessageType,
            messages: List[ChatCompletionMessageParam],
            model: str,
            message_id: str,
            thread_id: str,
            reply_fn: Callable[[Union[CompletionReply, CompletionStreamChunk]], None],
            user_input: Optional[str] = None,
            response_format_info: Optional[ResponseFormatInfo] = None
    ) -> str:
        """
        Stream completions from the AI provider and return the accumulated response.
        Returns: The accumulated response string.
        """
        # Reset the last error
        self.last_error = None
        accumulated_response = ""
        last_message_content = str(messages[-1].get('content', '')) if messages else ""

        # Send initial acknowledgment
        reply_fn(CompletionReply(
            items=[
                CompletionItem(content="", isIncomplete=True, token=message_id)
            ],
            parent_id=message_id,
        ))

        try:
            # Handle Claude API calls
            if constants.CLAUDE_MODEL and not self.api_key:
                if self._anthropic_client is None:
                    if constants.CLAUDE_API_KEY:
                        self._anthropic_client = AnthropicClient(
                            api_key=constants.CLAUDE_API_KEY,
                            model=constants.CLAUDE_MODEL
                        )
                    else:
                        self._anthropic_client = AnthropicClient(api_key=None, model=constants.CLAUDE_MODEL)

                accumulated_response = await self._anthropic_client.stream_response(
                    messages=messages,
                    message_type=message_type,
                    message_id=message_id,
                    reply_fn=reply_fn
                )

            # Handle Gemini API calls
            elif constants.GEMINI_MODEL and not self.api_key:
                if self._gemini_client is None:
                    if constants.GEMINI_API_KEY:
                        self._gemini_client = GeminiClient(
                            api_key=constants.GEMINI_API_KEY,
                            model=model
                        )
                    else:
                        self._gemini_client = GeminiClient(api_key=None, model=constants.GEMINI_MODEL)

                accumulated_response = await self._gemini_client.stream_completions(
                    messages=messages,
                    message_id=message_id,
                    reply_fn=reply_fn,
                )

            # Handle OpenAI and other providers
            elif self._openai_client:
                accumulated_response = await self._openai_client.stream_completions(
                    message_type=message_type,
                    messages=messages,
                    model=model,
                    message_id=message_id,
                    thread_id=thread_id,
                    reply_fn=reply_fn,
                    user_input=user_input,
                    response_format_info=response_format_info
                )
            else:
                raise ValueError("No AI provider configured")

            # Log the successful completion
            log_ai_completion_success(
                key_type=USER_KEY if self.key_type == "user" else MITO_SERVER_KEY,
                message_type=message_type,
                last_message_content=last_message_content,
                response={"completion": accumulated_response},
                user_input=user_input or "",
                thread_id=thread_id
            )

            return accumulated_response

        except BaseException as e:
            self.last_error = CompletionError.from_exception(e)
            log(
                MITO_AI_COMPLETION_ERROR,
                params={
                    KEY_TYPE_PARAM: self.key_type,
                    'message_type': message_type.value,
                },
                error=e
            )
            # Send error message to client before raising
            reply_fn(CompletionStreamChunk(
                parent_id=message_id,
                chunk=CompletionItem(
                    content="",
                    isIncomplete=True,
                    error=CompletionItemError(
                        message=f"Failed to process completion: {e!r}"
                    ),
                    token=message_id,
                ),
                done=True,
                error=CompletionError.from_exception(e),
            ))
            raise

